/***********************************************************************\
*                              translate.d                              *
*                                                                       *
*                         Unlambda to D compiler                        *
*                                                                       *
*                              Version 0.1                              *
*                                                                       *
*           Intermediate code structure and parsing algorithm           *
>***********************************************************************<
* Copyright (C) 2006-2007 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.unlambda.translate;
import
	std.string,
	std.ctype,
	std.stream;

private bool[char] needOutput;
private bool[char] needCompare;


char[] parse(inout char[] code) {
	while (true) {  // loop until a portion of code has actually been read
		if (code.length == 0) {
			throw new Exception("Expected instruction, but found EOF");
		}
		char instr = code[0];
		code = code[1..$];

		switch (instr) {
			case '`':
				// first, handle special cases
				if (code.length >= 4 && code[0..4] == "`sii") {
					code = code[4..$];
					return "SII";
				}
				if (code.length >= 2
					  && find("sk", code[0]) != -1
					  && find("ski", code[1]) != -1) {
					scope(exit) code = code[2..$];
					return std.string.toupper(code[0..2]);
				}

				char[] operator = parse(code);
				char[] operand = parse(code);
				return operator ~ '(' ~ operand ~ ')';

			case 's', 'k', 'i', 'v', 'c', 'd', 'e', 'r':
				return "" ~ cast(char) std.ctype.toupper(instr);

			case '.':
				if (code.length == 0) {
					throw new Exception(
					  "Expected character after '.', but found EOF");
				}
				char c = code[0];

				code = code[1..$];
				if (c == '\n') {
					return "R";
				} else {
					needOutput[c] = true;
					return format("_%d", c);
				}

			case '?':
				if (code.length == 0) {
					throw new Exception(
					  "Expected character after '?', but found EOF");
				}
				char c = code[0];

				code = code[1..$];
				needCompare[c] = true;
				return format("_c%d", c);

			case '@':
				return "IB";

			case '|':
				return "RB";

			case ' ', '\t', '\n':
				break;

			case '#':
				int eol = find(code, '\n');
				if (eol == -1) {
					throw new Exception("Expected instruction, but found EOF");
				}
				code = code[eol+1 .. $];
				break;

			default:
				throw new Exception(
				  "Expected instruction, but found '" ~ instr ~ "'");
		}
	}
}


void writeCharsNeeded(OutputStream output) {
	if (needOutput.length != 0) {
		bool atStart = true;

		output.writefln("\tOutCharFunction");

		foreach (char c, bool b; needOutput) {
			assert (b);
			if (!atStart) output.writefln(',');
			output.writef("\t\t_%d = OutCharFunction['%s']", c, c);
			atStart = false;
		}
		output.writefln(';');
	}

	if (needCompare.length != 0) {
		bool atStart = true;

		output.writefln("\tCompareBufFunction");

		foreach (char c, bool b; needCompare) {
			assert (b);
			if (!atStart) output.writefln(',');
			output.writef("\t\t_%d = CompareBufFunction['%s']", c, c);
			atStart = false;
		}
		output.writefln(";");
	}
}
